import numpy as np
from scipy import stats

try:
    from dyfunconn import fc
except ImportError:
    fc = None


def get_fc_graphs(x, fc_measure, nf_mode,
                  samples_per_graph=None, link_cutoff=0., percentiles=None,
                  self_loops=True, band_freq=None, band_freq_hi=None,
                  sampling_freq=None, nfft=None, n_overlap=None):
    """
    Compute a sequence of functional connectivity networks from the given time
    series. Each graph will be computed using consecutive windows of
    `samples_per_graph` timesteps. If x.shape[1] % samples_per_graph != 0, then
    the remaining time steps at the end of the time series are discarded.
    :param x: np.array of shape (n_channels, n_timesteps)
    :param fc_measure: string, the FC measure to use (see _get_fc_graph())
    :param nf_mode: string, how to compute node features (see _get_fc_graph())
    :param samples_per_graph: int, number of samples used to compute a graph.
    :param link_cutoff: remove links in the FC graphs if the absolute value of
    their FC measure is smaller than the threshold. Overrides `percentiles` flag.
    :param percentiles: tuple of 2 floats, keep only links whose FC measure is
    below the `percentiles[0]`-th percentile or above the `percentiles[1]`-th
    percentile.
    :param self_loops: add self-loops to the adjacency matrices.
    :param band_freq: tuple of 2 floats >0, filter the signal with a bandpass
    filter in this band. mandatory when using dyfunconn.
    :param band_freq_hi: mandatory when using dyfunconn.aec().
    :param sampling_freq: sampling frequency of the data. Mandatory when
    using dyfunconn.
    :param nfft: mandatory when using dyfunconn.wpli() and dyfunconn.dwpli().
    :param n_overlap: mandatory when using dyfunconn.wpli() and dyfunconn.dwpli().
    :return:
    """
    if fc is None:
        raise ImportError('Creating functional connectivity networks '
                          'requires dyfunconn.')
    if x.ndim != 2:
        raise ValueError('Expected x to have rank 2, got {}'.format(x.ndim))
    if samples_per_graph is None:
        samples_per_graph = x.shape[1]

    output = _loop(
        x, samples_per_graph, fc_measure, nf_mode,
        band_freq=band_freq, band_freq_hi=band_freq_hi, sampling_freq=sampling_freq,
        nfft=nfft, n_overlap=n_overlap
    )

    nf, ef = zip(*output)
    nf = np.array(nf)
    ef = np.array(ef)

    # Compute adjacency matrix
    adj = np.ones_like(ef)
    if link_cutoff != 0:
        adj[np.abs(adj) < link_cutoff] = 0
    elif percentiles is not None:
        N = adj.shape[-1]
        pctl_lo = []
        pctl_hi = []
        for fcm in ef.reshape((-1, N * N)).T:
            pctl_lo.append(stats.scoreatpercentile(fcm, percentiles[0]))
            pctl_hi.append(stats.scoreatpercentile(fcm, percentiles[1]))
        pctl_lo = np.array(pctl_lo).reshape((-1, N, N))
        pctl_hi = np.array(pctl_hi).reshape((-1, N, N))
        adj = np.logical_or(ef < pctl_lo, ef > pctl_hi).astype(np.int)
    else:
        pass

    if self_loops:
        for i in range(adj.shape[0]):
            np.fill_diagonal(adj[i], 1.0)
    else:
        # Set the main diagonal to zero
        for i in range(adj.shape[0]):
            np.fill_diagonal(adj[i], 0.0)
            np.fill_diagonal(ef[i], 0.0)

    nf = nf[..., None]
    ef = ef[..., None]

    return adj, nf, ef


def _loop(x, samples_per_graph, fc_measure='corr', nf_mode='mean',
          band_freq=None, band_freq_hi=None, sampling_freq=None, nfft=None,
          n_overlap=None):
    output = [
        _get_fc_graph(
            x[:, i: i + samples_per_graph], fc_measure, nf_mode,
            band_freq=band_freq, band_freq_hi=band_freq_hi,
            sampling_freq=sampling_freq, nfft=nfft, n_overlap=n_overlap)
        for i in range(0, x.shape[1], samples_per_graph)
    ]
    return output


def _get_fc_graph(x, fc_measure, nf_mode, band_freq=None, band_freq_hi=None,
                  sampling_freq=None, nfft=None, n_overlap=None):
    if fc_measure == 'iplv':
        _, ef = fc.iplv(x, band_freq, sampling_freq)
    elif fc_measure == 'icoh':
        ef = fc.icoherence(x, band_freq, sampling_freq)
    elif fc_measure == 'corr':
        ef = fc.corr(x, band_freq, sampling_freq)
    elif fc_measure == 'aec':
        ef = fc.aec(x, band_freq, band_freq_hi, sampling_freq)
    elif fc_measure == 'wpli':
        csdparams = {'NFFT': nfft, 'noverlap': n_overlap}
        ef = fc.wpli(x, band_freq, sampling_freq, **csdparams)
    elif fc_measure == 'dwpli':
        csdparams = {'NFFT': nfft, 'noverlap': n_overlap}
        ef = fc.dwpli(x, band_freq, sampling_freq, **csdparams)
    elif fc_measure == 'dpli':
        ef = fc.dpli(x, band_freq, sampling_freq)
    else:
        raise ValueError('Invalid fc_measure ' + fc_measure)

    # Dummy node features
    if nf_mode == 'full':
        nf = x.copy()
    elif nf_mode == 'mean':
        nf = np.mean(x, -1)
    elif nf_mode == 'energy':
        nf = np.sum(x ** 2, -1)
    elif nf_mode == 'power':
        nf = np.sum(x ** 2, -1) / x.shape[-1]
    elif nf_mode == 'ones':
        nf = np.ones((ef.shape[0], 1))
    else:
        raise ValueError('Invalid nf_mode' + nf_mode)
    return nf, ef



from spektral.layers.ops import matmul_A_B, matmul_AT_B, matmul_A_BT, matmul_AT_B_A, sp_matrix_to_sp_tensor, reshape
import numpy as np
from keras import backend as K

sess = K.get_session()
batch_size = 10
N = 3
tol = 5.e-7


def _assert_all_close(output, expected_output):
    try:
        assert np.allclose(output, expected_output, atol=tol)
    except AssertionError:
        mean_diff = np.mean(np.absolute(output - expected_output))
        max_diff = np.max(np.absolute(output - expected_output))
        if mean_diff <= tol:
            print('Max difference above tolerance, but mean OK.')
        else:
            raise AssertionError('Mean diff: {}, Max diff: {}'.format(
                mean_diff, max_diff
            ))


def _convert_to_sparse_tensor(x):
    if x.ndim == 2:
        return sp_matrix_to_sp_tensor(x)
    elif x.ndim == 3:
        s1_, s2_, s3_ = x.shape
        return reshape(
            sp_matrix_to_sp_tensor(x.reshape(s1_ * s2_, s3_)),
            (s1_, s2_, s3_)
        )


def _cast_all_to_dtype(values, dtype):
    return [K.cast(v, dtype) for v in values]


def _check_op(op, numpy_inputs, expected_output, convert_to_sparse=None):
    output = _check_op_dense(op, numpy_inputs)
    _assert_all_close(output, expected_output)

    if convert_to_sparse:
        if isinstance(convert_to_sparse, list):
            if isinstance(convert_to_sparse[0], bool):
                # Make it into a list of list, always
                convert_to_sparse = [convert_to_sparse]

        for c_t_s in convert_to_sparse:
            output = _check_op_sparse(op, numpy_inputs, c_t_s)
            _assert_all_close(output, expected_output)


def _check_op_dense(op, numpy_inputs):
    tf_inputs = [K.constant(x) for x in numpy_inputs]
    tf_inputs = _cast_all_to_dtype(tf_inputs, np.float32)

    op_result = op(*tf_inputs)
    output = sess.run(op_result)

    return np.asarray(output)


def _check_op_sparse(op, numpy_inputs, convert_to_sparse):
    tf_inputs = []
    for i in range(len(numpy_inputs)):
        if convert_to_sparse[i]:
            tf_inputs.append(
                _convert_to_sparse_tensor(numpy_inputs[i])
            )
        else:
            tf_inputs.append(
                K.constant(numpy_inputs[i])
            )
    tf_inputs = _cast_all_to_dtype(tf_inputs, np.float32)

    op_result = op(*tf_inputs)
    output = sess.run(op_result)

    if hasattr(output, 'toarray'):
        return output.toarray()
    else:
        return np.asarray(output)


def test_matmul_ops_single_mode():
    A = np.random.randn(N, N)
    B = np.random.randn(N, N)
    convert_to_sparse = [[True, False], [False, True]]
    _check_op(matmul_A_B, [A, B], A.dot(B), convert_to_sparse)
    _check_op(matmul_AT_B_A, [A, B], A.T.dot(B).dot(A), convert_to_sparse)
    _check_op(matmul_AT_B, [A, B], A.T.dot(B), convert_to_sparse)
    _check_op(matmul_A_BT, [A, B], A.dot(B.T), convert_to_sparse)


def test_matmul_ops_mixed_mode():
    A = np.random.randn(N, N)
    B = np.random.randn(batch_size, N, N)
    convert_to_sparse = [[True, False], [False, True]]

    # A * B
    expected_output = np.array([A.dot(B[i]) for i in range(batch_size)])
    _check_op(matmul_A_B, [A, B], expected_output, convert_to_sparse)

    # A.T * B * A
    expected_output = np.array([A.T.dot(B[i]).dot(A) for i in range(batch_size)])
    _check_op(matmul_AT_B_A, [A, B], expected_output, convert_to_sparse)

    # A.T * B
    expected_output = np.array([A.T.dot(B[i]) for i in range(batch_size)])
    _check_op(matmul_AT_B, [A, B], expected_output, convert_to_sparse)

    # A * B.T
    expected_output = np.array([A.dot(B[i].T) for i in range(batch_size)])
    _check_op(matmul_A_BT, [A, B], expected_output, convert_to_sparse)


def test_matmul_ops_inv_mixed_mode():
    A = np.random.randn(batch_size, N, N)
    B = np.random.randn(N, N)

    # A * B
    expected_output = np.array([A[i].dot(B) for i in range(batch_size)])
    _check_op(matmul_A_B, [A, B], expected_output)

    # A.T * B * A
    expected_output = np.array([A[i].T.dot(B).dot(A[i]) for i in range(batch_size)])
    _check_op(matmul_AT_B_A, [A, B], expected_output)

    # A.T * B
    expected_output = np.array([A[i].T.dot(B) for i in range(batch_size)])
    _check_op(matmul_AT_B, [A, B], expected_output)

    # A * B.T
    expected_output = np.array([A[i].dot(B.T) for i in range(batch_size)])
    _check_op(matmul_A_BT, [A, B], expected_output)


def test_matmul_ops_batch_mode():
    A = np.random.randn(batch_size, N, N)
    B = np.random.randn(batch_size, N, N)

    # A * B
    expected_output = np.array([A[i].dot(B[i]) for i in range(batch_size)])
    _check_op(matmul_A_B, [A, B], expected_output)

    # A.T * B * A
    expected_output = np.array([A[i].T.dot(B[i]).dot(A[i]) for i in range(batch_size)])
    _check_op(matmul_AT_B_A, [A, B], expected_output)

    # A.T * B
    expected_output = np.array([A[i].T.dot(B[i]) for i in range(batch_size)])
    _check_op(matmul_AT_B, [A, B], expected_output)

    # A * B.T
    expected_output = np.array([A[i].dot(B[i].T) for i in range(batch_size)])
    _check_op(matmul_A_BT, [A, B], expected_output)



import numpy as np
from keras import activations, initializers, regularizers, constraints
from keras import backend as K
from keras.layers import Layer


class InnerProduct(Layer):
    """
    Computes the inner product between elements of a given 2d tensor \(x\): 
    $$
        \\langle x, x \\rangle = xx^T.
    $$

    **Mode**: single.

    **Input**

    - rank 2 tensor of shape `(input_dim_1, input_dim_2)` (e.g. node features
    of shape `(num_nodes, num_features)`);

    **Output**

    - rank 2 tensor of shape `(input_dim_1, input_dim_1)`

    :param trainable_kernel: add a trainable square matrix between the inner
    product (i.e., `x.dot(w).dot(x.T)`);
    :param activation: activation function to use;
    :param kernel_initializer: initializer for the kernel matrix;
    :param kernel_regularizer: regularization applied to the kernel;
    :param activity_regularizer: regularization applied to the output;
    :param kernel_constraint: constraint applied to the kernel;
    """
    def __init__(self,
                 trainable_kernel=False,
                 activation=None,
                 kernel_initializer='glorot_uniform',
                 kernel_regularizer=None,
                 activity_regularizer=None,
                 kernel_constraint=None,
                 **kwargs):
        if 'input_shape' not in kwargs and 'input_dim' in kwargs:
            kwargs['input_shape'] = (kwargs.pop('input_dim'),)
        super(InnerProduct, self).__init__(**kwargs)
        self.trainable_kernel = trainable_kernel
        self.activation = activations.get(activation)
        self.kernel_initializer = initializers.get(kernel_initializer)
        self.kernel_regularizer = regularizers.get(kernel_regularizer)
        self.activity_regularizer = regularizers.get(activity_regularizer)
        self.kernel_constraint = constraints.get(kernel_constraint)

    def build(self, input_shape):
        assert len(input_shape) >= 2
        if self.trainable_kernel:
            features_dim = input_shape[-1]
            self.kernel = self.add_weight(shape=(features_dim, features_dim),
                                          initializer=self.kernel_initializer,
                                          name='kernel',
                                          regularizer=self.kernel_regularizer,
                                          constraint=self.kernel_constraint)
        self.built = True

    def call(self, inputs):
        if self.trainable_kernel:
            output = K.dot(K.dot(inputs, self.kernel), K.transpose(inputs))
        else:
            output = K.dot(inputs, K.transpose(inputs))
        if self.activation is not None:
            output = self.activation(output)
        return output

    def compute_output_shape(self, input_shape):
        if len(input_shape) == 2:
            return (None, None)
        else:
            return input_shape[:-1] + (input_shape[-2], )

    def get_config(self, **kwargs):
        config = {}
        base_config = super(InnerProduct, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))


class MinkowskiProduct(Layer):
    """
    Computes the hyperbolic inner product between elements of a given 2d tensor
    \(x\): 
    $$
        \\langle x, x \\rangle = x \,
        \\begin{pmatrix}
            I_{d\\times d} & 0 \\\\ 0 & -1
        \\end{pmatrix} \\,x^T.
    $$

    **Mode**: single.

    **Input**

    - rank 2 tensor of shape `(input_dim_1, input_dim_2)` (e.g. node features
    of shape `(num_nodes, num_features)`);

    **Output**

    - rank 2 tensor of shape `(input_dim_1, input_dim_1)`

    :param input_dim_1: first dimension of the input tensor; set this if you
    encounter issues with shapes in your model, in order to provide an explicit
    output shape for your layer.
    :param activation: activation function to use;
    :param activity_regularizer: regularization applied to the output;
    """
    def __init__(self,
                 input_dim_1=None,
                 activation=None,
                 activity_regularizer=None,
                 **kwargs):
        if 'input_shape' not in kwargs and 'input_dim' in kwargs:
            kwargs['input_shape'] = (kwargs.pop('input_dim'),)
        super(MinkowskiProduct, self).__init__(**kwargs)
        self.input_dim_1 = input_dim_1
        self.activation = activations.get(activation)
        self.activity_regularizer = regularizers.get(activity_regularizer)

    def build(self, input_shape):
        assert len(input_shape) >= 2
        self.built = True

    def call(self, inputs):
        F = K.int_shape(inputs)[-1]
        minkowski_prod_mat = np.eye(F)
        minkowski_prod_mat[-1, -1] = -1.
        minkowski_prod_mat = K.constant(minkowski_prod_mat)
        output = K.dot(inputs, minkowski_prod_mat)
        output = K.dot(output, K.transpose(inputs))
        output = K.clip(output, -10e9, -1.)

        if self.activation is not None:
            output = self.activation(output)

        return output

    def compute_output_shape(self, input_shape):
        if len(input_shape) == 2:
            if self.input_dim_1 is None:
                return (None, None)
            else:
                return (self.input_dim_1, self.input_dim_1)
        else:
            return input_shape[:-1] + (input_shape[-2], )

    def get_config(self, **kwargs):
        config = {}
        base_config = super(MinkowskiProduct, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))



import networkx as nx

from spektral.utils.conversion import numpy_to_nx


def plot_nx(nx_graph, nf_name=None, ef_name=None, layout='spring_layout',
            labels=True, node_color='r', node_size=300, **kwds):
    """
    Plot the given Networkx graph.
    :param nx_graph: a Networkx graph
    :param nf_name: name of the relevant node feature to plot
    :param ef_name: name of the relevant edgee feature to plot
    :param layout: type of layout for networkx
    :param labels: plot labels
    :param node_color: color for the plotted nodes
    :param node_size: size of the plotted nodes
    :return: None
    """
    layout = _deserialize_nx_layout(layout, nf_name=nf_name)
    pos = layout(nx_graph)
    nx.draw(nx_graph, pos, node_color=node_color, node_size=node_size, **kwds)
    if nf_name is not None:
        node_labels = nx.get_node_attributes(nx_graph, nf_name)
        if labels:
            nx.draw_networkx_labels(nx_graph, pos, labels=node_labels)
    if ef_name is not None:
        edge_labels = nx.get_edge_attributes(nx_graph, ef_name)
        if labels:
            nx.draw_networkx_edge_labels(nx_graph, pos, edge_labels=edge_labels)


def plot_numpy(adj, node_features=None, edge_features=None, nf_name=None,
               ef_name=None, layout='spring_layout', labels=True,
               node_color='r', node_size=300, ):
    """
    Converts a graph in matrix format (i.e. with adjacency matrix, node features
    matrix, and edge features matrix) to the Networkx format, then plots it with
    plot_nx().
    :param adj: np.array, adjacency matrix of the graph 
    :param node_features: np.array, node features matrix of the graph
    :param edge_features: np.array, edge features matrix of the graph
    :param nf_name: name to assign to the node features
    :param ef_name: name to assign to the edge features
    :param layout: type of layout for networkx
    :param labels: plot labels
    :param node_color: color for the plotted nodes
    :param node_size: size of the plotted nodes
    :return: None
    """
    if node_features is not None and nf_name is None:
        nf_name = 'nf'
    if edge_features is not None and ef_name is None:
        ef_name = 'ef'
    g = numpy_to_nx(adj, node_features, edge_features, nf_name, ef_name)
    plot_nx(g, nf_name, ef_name, node_color=node_color, node_size=node_size, layout=layout, labels=labels)


# Utils
def _coordinates_layout_closure(nf_name):
    def coordinates_layout(nx_graph):
        return nx.get_node_attributes(nx_graph, nf_name)

    return coordinates_layout


def _deserialize_nx_layout(layout, nf_name=None):
    if isinstance(layout, str):
        if layout in nx.layout.__all__:
            return eval('nx.{}'.format(layout))
        elif layout == 'coordinates':
            if nf_name is None:
                raise ValueError('nf_name cannot be None')
            return _coordinates_layout_closure(nf_name)
        else:
            raise ValueError('layout must be in nx.layout.__all__ or \'coordinates\'')



import os
import shutil
import zipfile

import networkx as nx
import numpy as np
import pandas as pd
import requests
from sklearn.preprocessing import OneHotEncoder, StandardScaler

from spektral.utils import nx_to_numpy

DATASET_URL = 'https://ls11-www.cs.tu-dortmund.de/people/morris/graphkerneldatasets'
DATASET_CLEAN_URL = 'https://raw.githubusercontent.com/nd7141/graph_datasets/master/datasets'
DATA_PATH = os.path.expanduser('~/.spektral/datasets/')
AVAILABLE_DATASETS = [
    d[:-4]
    for d in pd.read_html(DATASET_URL)[0].Name[2:-1].values.tolist()
]


def load_data(dataset_name, normalize_features='zscore', clean=False):
    """
    Loads one of the Benchmark Data Sets for Graph Kernels from TU Dortmund
    ([link](https://ls11-www.cs.tu-dortmund.de/staff/morris/graphkerneldatasets)).
    The node features are computed by concatenating the following features for
    each node:

    - node attributes, if available, normalized as specified in `normalize_features`;
    - clustering coefficient, normalized with z-score;
    - node degrees, normalized as specified in `normalize_features`;
    - node labels, if available, one-hot encoded.
    :param dataset_name: name of the dataset to load (see `spektral.datasets.tud.AVAILABLE_DATASETS`).
    :param normalize_features: `'zscore'` or `'ohe'`, how to normalize the node
    features (only works for node attributes and node degrees).
    :param clean: if True, return a version of the dataset with no isomorphic
    graphs.
    :return:
    - a list of adjacency matrices;
    - a list of node feature matrices;
    - a numpy array containing the one-hot encoded targets.
    """
    if dataset_name not in AVAILABLE_DATASETS:
        raise ValueError('Available datasets: {}'.format(AVAILABLE_DATASETS))

    if clean:
        dataset_name += '_clean'
    if not os.path.exists(DATA_PATH + dataset_name):
        _download_data(dataset_name)

    # Read data
    nx_graphs, y = _read_graphs(dataset_name)

    # Preprocessing
    y = np.array(y)[..., None]
    y = OneHotEncoder(sparse=False, categories='auto').fit_transform(y)

    # Get node attributes
    try:
        A, X_attr, _ = nx_to_numpy(nx_graphs, nf_keys=['attributes'], auto_pad=False)
        X_attr = _normalize_node_features(X_attr, normalize_features)
    except KeyError:
        print('Featureless nodes')
        A, X_attr, _ = nx_to_numpy(nx_graphs, auto_pad=False)  # na will be None

    # Get clustering coefficients (always zscore norm)
    clustering_coefficients = [np.array(list(nx.clustering(g).values()))[..., None] for g in nx_graphs]
    clustering_coefficients = _normalize_node_features(clustering_coefficients, 'zscore')

    # Get node degrees
    node_degrees = np.array([np.sum(_, axis=-1, keepdims=True) for _ in A])
    node_degrees = _normalize_node_features(node_degrees, normalize_features)

    # Get node labels (always ohe norm)
    try:
        _, X_labs, _ = nx_to_numpy(nx_graphs, nf_keys=['label'], auto_pad=False)
        X_labs = _normalize_node_features(X_labs, 'ohe')
    except KeyError:
        print('Label-less nodes')
        X_labs = None

    # Concatenate features
    Xs = [node_degrees, clustering_coefficients]
    if X_attr is not None:
        Xs.append(X_attr)
    if X_labs is not None:
        Xs.append(X_labs)
    X = [np.concatenate(x_, axis=-1) for x_ in zip(*Xs)]
    X = np.array(X)

    return A, X, y


def _read_graphs(dataset_name):
    file_prefix = DATA_PATH + dataset_name + '/' + dataset_name
    with open(file_prefix + "_graph_indicator.txt", "r") as f:
        graph_indicator = [int(i) - 1 for i in list(f)]

    # Nodes
    num_graphs = max(graph_indicator)
    node_indices = []
    offset = []
    c = 0

    for i in range(num_graphs + 1):
        offset.append(c)
        c_i = graph_indicator.count(i)
        node_indices.append((c, c + c_i - 1))
        c += c_i

    graph_list = []
    vertex_list = []
    for i in node_indices:
        g = nx.Graph(directed=False)
        vertex_list_g = []
        for j in range(i[1] - i[0] + 1):
            vertex_list_g.append(g.add_node(j))

        graph_list.append(g)
        vertex_list.append(vertex_list_g)

    # Edges
    with open(file_prefix + "_A.txt", "r") as f:
        edges = [i.split(',') for i in list(f)]

    edges = [(int(e[0].strip()) - 1, int(e[1].strip()) - 1) for e in edges]

    edge_indicator = []
    edge_list = []
    for e in edges:
        g_id = graph_indicator[e[0]]
        edge_indicator.append(g_id)
        g = graph_list[g_id]
        off = offset[g_id]

        # Avoid multigraph
        edge_list.append(g.add_edge(e[0] - off, e[1] - off))

    # Node labels
    if os.path.exists(file_prefix + "_node_labels.txt"):
        with open(file_prefix + "_node_labels.txt", "r") as f:
            node_labels = [int(i) for i in list(f)]

        i = 0
        for g in graph_list:
            for n in g.nodes():
                g.nodes[n]['label'] = node_labels[i]
                i += 1

    # Node Attributes
    if os.path.exists(file_prefix + "_node_attributes.txt"):
        with open(file_prefix + "_node_attributes.txt", "r") as f:
            node_attributes = [map(float, i.split(',')) for i in list(f)]
        i = 0
        for g in graph_list:
            for n in g.nodes():
                g.nodes[n]['attributes'] = list(node_attributes[i])
                i += 1

    # Classes
    with open(file_prefix + "_graph_labels.txt", "r") as f:
        classes = [int(i) for i in list(f)]

    return graph_list, classes


def _download_data(dataset_name):
    print('Dowloading ' + dataset_name + ' dataset.')
    if dataset_name.endswith('_clean'):
        true_name = dataset_name[:-6]
        url = DATASET_CLEAN_URL
    else:
        true_name = dataset_name
        url = DATASET_URL

    data_url = '{}/{}.zip'.format(url, true_name)
    req = requests.get(data_url)
    with open(DATA_PATH + dataset_name + '.zip', 'wb') as out_file:
        out_file.write(req.content)
    with zipfile.ZipFile(DATA_PATH + dataset_name + '.zip', 'r') as zip_ref:
        zip_ref.extractall(DATA_PATH + dataset_name + '/')
    os.remove(DATA_PATH + dataset_name + '.zip')

    subfolder = os.path.join(DATA_PATH, dataset_name, true_name)
    parentfolder = os.path.join(DATA_PATH, dataset_name)
    for filename in os.listdir(subfolder):
        try:
            suffix = filename.split(true_name)[1]
        except IndexError:
            # Probably the README
            continue
        shutil.move(
            os.path.join(subfolder, filename),
            os.path.join(parentfolder, dataset_name + suffix)
        )
    shutil.rmtree(subfolder)


def _normalize_node_features(feat_list, norm='ohe'):
    """
    Apply one-hot encoding or z-score to a list of node features
    """
    if norm == 'ohe':
        fnorm = OneHotEncoder(sparse=False, categories='auto')
    elif norm == 'zscore':
        fnorm = StandardScaler()
    else:
        raise ValueError('Possible feat_norm: ohe, zscore')
    fnorm.fit(np.vstack(feat_list))
    feat_list = [fnorm.transform(feat_.astype(np.float32)) for feat_ in feat_list]
    return feat_list
